
#include "SDL/sdl.h"
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#define FBIO_MAGIC 'D'
#define FBIO_LCD_CHANGE_CONTROL _IOW(FBIO_MAGIC, 90, unsigned int[2])
#define LCD_DIRECTION_ON_CMD 5 /* 320x240 */
#define LCD_DIRECTION_OFF_CMD 6 /* 240x320 */

#define UP_MASK 	0x83
#define DOWN_MASK 	0x38
#define LEFT_MASK 	0x0E
#define RIGHT_MASK 	0xE0

#if WIZ
SDL_Surface *WIZ_SetVideoMode(int width, int height, int bitsperpixel, Uint32 flags){
	SDL_Surface *s;
	s = SDL_SetVideoMode(width, height, bitsperpixel, flags);
	if(s){
		unsigned int send[2];int fb_fd = open("/dev/fb0", O_RDWR);
		send[0] = LCD_DIRECTION_OFF_CMD;
		ioctl(fb_fd, FBIO_LCD_CHANGE_CONTROL, &send);
		close(fb_fd);
	}
	return s;
}
#endif

int GP2X_PollEvent(SDL_Event *ev){

	static unsigned int joy_status = 0;
	static signed int joy_x=0, joy_y=0;
	SDL_Event event;
	SDL_Event jevent;
	unsigned int status;

	status = joy_status;
	if (SDL_PollEvent(&event)) switch(event.type){
		case SDL_JOYBUTTONUP:
			if(event.jbutton.which==0 && event.jbutton.button<8){
				status &= ~(1<<event.jbutton.button);
				break;
			}
			*ev = event;
			return 1;
		case SDL_JOYBUTTONDOWN:
			if(event.jbutton.which==0 && event.jbutton.button<8){
				status |= 1<<event.jbutton.button;
				break;
			}
			*ev = event;
			return 1;
		default:
			*ev = event;
			return 1;
	} else {
		return 0;
	}

	if(status==joy_status) return 0;

	jevent.type = SDL_JOYAXISMOTION;
	jevent.jaxis.which = 0;

	jevent.jaxis.axis = 1;
	jevent.jaxis.value = 0;
	jevent.jaxis.value += status & UP_MASK ? -32768 : 0;
	jevent.jaxis.value += status & DOWN_MASK ? 32767 : 0;
	if(joy_y != jevent.jaxis.value) SDL_PushEvent(&jevent);
	joy_y = jevent.jaxis.value;

	jevent.jaxis.axis = 0;
	jevent.jaxis.value = 0;
	jevent.jaxis.value += status & LEFT_MASK ? -32768 : 0;
	jevent.jaxis.value += status & RIGHT_MASK ? 32767 : 0;
	if(joy_x != jevent.jaxis.value) SDL_PushEvent(&jevent);
	joy_x = jevent.jaxis.value;

	joy_status = status;

	*ev = event;

	return 1;
}
